# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import ListMixin

__all__ = [
    "Activity",
    "ActivityManager",
]


class Activity(RestObject):
    pass


class ActivityManager(RestManager, ListMixin):
    _path = "activities"
    _obj_cls = Activity



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, UpdateMixin

__all__ = [
    "FindingComment",
    "FindingCommentManager",
    "FindingFindingCommentManager",
]


class FindingComment(RestObject):
    pass


class FindingCommentManager(RestManager, UpdateMixin, DeleteMixin):
    _path = "finding-comments"
    _obj_cls = FindingComment


class FindingFindingCommentManager(RestManager, CreateMixin):
    _path = "findings/{finding_id}/finding-comments"
    _parent_attrs = {"finding_id": "id"}
    _obj_cls = FindingComment



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, UpdateMixin

__all__ = [
    "FindingRetestInquiry",
    "FindingRetestInquiryManager",
    "FindingFindingRetestInquiryManager",
]


class FindingRetestInquiry(RestObject):
    pass


class FindingRetestInquiryManager(RestManager, UpdateMixin, DeleteMixin):
    _path = "finding-retest-inquiries"
    _obj_cls = FindingRetestInquiry


class FindingFindingRetestInquiryManager(RestManager, CreateMixin):
    _path = "findings/{finding_id}/finding-retest-inquiries"
    _parent_attrs = {"finding_id": "id"}
    _obj_cls = FindingRetestInquiry



# pylint: disable = missing-module-docstring, missing-class-docstring

from typing import TYPE_CHECKING, Optional, List, Mapping
from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, GetMixin, ListMixin, UpdateMixin

__all__ = [
    "User",
    "UserManager",
]


class User(RestObject):
    pass


class UserManager(RestManager, CreateMixin, GetMixin, ListMixin, UpdateMixin):
    _path = "users"
    _obj_cls = User

    def me(
        self,
        include: Optional[List[str]] = None,
        query_data: Optional[Mapping[str, str]] = None,
        **kwargs,
    ) -> User:
        """Get the user who owns the API token

        Args:
            include: Related data to include in the response.
            query_data: Dict of additional query parameters
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The response from the server, serialized into the `User` type.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """

        query_data = dict(query_data) if query_data else {}

        if include:
            query_data["include"] = ",".join(include)

        result = self.reporter.http_request(
            verb="get",
            path="users/me",
            query_data=query_data,
            **kwargs,
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return User(self.reporter, result.json())



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin

__all__ = [
    "Webhook",
    "WebhookManager",
]


class Webhook(RestObject):
    pass


class WebhookManager(RestManager, CreateMixin, DeleteMixin):
    _path = "webhooks"
    _obj_cls = Webhook



# pylint: disable = missing-module-docstring, missing-class-docstring, redefined-builtin

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, UpdateMixin, DeleteMixin

__all__ = [
    "AssessmentComment",
    "AssessmentCommentManager",
    "AssessmentAssessmentCommentManager",
    "AssessmentCommentReplyManager",
]


class AssessmentComment(RestObject):
    pass


class AssessmentCommentManager(RestManager, UpdateMixin, DeleteMixin):
    _path = "assessment-comments"
    _obj_cls = AssessmentComment


class AssessmentAssessmentCommentManager(RestManager, CreateMixin):
    _path = "assessments/{assessment_id}/assessment-comments"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = AssessmentComment


class AssessmentCommentReplyManager(RestManager, CreateMixin):
    _path = "assessment-events/{assessment_event_id}/assessment-comments"
    _parent_attrs = {"assessment_event_id": "id"}
    _obj_cls = AssessmentComment



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, GetMixin, ListMixin, UpdateMixin


__all__ = [
    "Task",
    "TaskManager",
    "AssessmentTaskManager",
]


class Task(RestObject):
    pass


class TaskManager(RestManager, DeleteMixin, GetMixin, ListMixin, UpdateMixin):
    _path = "tasks"
    _obj_cls = Task


class AssessmentTaskManager(RestManager, CreateMixin):
    _path = "assessments/{assessment_id}/tasks"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = Task



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, GetMixin, ListMixin, UpdateMixin

__all__ = [
    "Target",
    "TargetManager",
    "AssessmentTargetManager",
]


class Target(RestObject):
    pass


class TargetManager(RestManager, DeleteMixin, GetMixin, ListMixin, UpdateMixin):
    _path = "targets"
    _obj_cls = Target


class AssessmentTargetManager(RestManager, CreateMixin):
    _path = "assessments/{assessment_id}/targets"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = Target



"""Mixins for model CRUD operations.

Instances of :class:`~reporter.base.RestManager` should derive these mixins according to the
operations possible on their corresponding :class:`~reporter.base.RestObject` in the Reporter API.

"""

# pylint: disable = too-few-public-methods, redefined-builtin, invalid-name

from typing import (
    Any,
    Callable,
    Generic,
    List,
    Mapping,
    Optional,
    Type,
    TYPE_CHECKING,
    TypeVar,
)

from reporter.base import RestList, RestManager, RestObject
from reporter.client import Reporter


__all__ = [
    "CreateMixin",
    "CrudMixin",
    "DeleteMixin",
    "GetMixin",
    "ListMixin",
    "SearchMixin",
    "UpdateMixin",
]


ChildOfRestObject = TypeVar("ChildOfRestObject", bound=RestObject)


class CreateMixin(Generic[ChildOfRestObject]):
    """Manager can create object."""

    _path: str
    _obj_cast: Callable
    _obj_cls: Type[ChildOfRestObject]
    reporter: Reporter

    def create(
        self,
        attrs: Mapping[str, Any],
        file: Optional[Any] = None,
        **kwargs: Any,
    ) -> ChildOfRestObject:
        """Create a new object.

        Args:
            attrs: Attributes for the created object.
            file: A file to upload when creating the object, if any.
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The response from the server, serialized into the object type.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """

        files = {"file": file} if file is not None else None

        result = self.reporter.http_request(
            verb="post",
            path=self._path,
            post_data=attrs,
            files=files,
            **kwargs,
        )

        if TYPE_CHECKING:
            assert isinstance(self, RestManager)

        if hasattr(self, "_obj_cast"):
            return self._obj_cast(self.reporter, result.json())
        return self._obj_cls(self.reporter, result.json())


class DeleteMixin(Generic[ChildOfRestObject]):
    """Manager can delete object."""

    _path: str
    _obj_cls: Type[ChildOfRestObject]
    _obj_cast: Callable
    reporter: Reporter

    def delete(
        self,
        id: str,
        **kwargs: Any,
    ):
        """Delete an object.

        Args:
            id: The ID of the object to delete.
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """
        path = f"{self._path}/{id}"

        self.reporter.http_request(
            verb="delete",
            path=path,
            **kwargs,
        )


class GetMixin(Generic[ChildOfRestObject]):
    """Manager can retrieve object."""

    _path: str
    _obj_cls: Type[ChildOfRestObject]
    _obj_cast: Callable
    reporter: Reporter

    def get(
        self,
        id: str,
        include: Optional[List[str]] = None,
        query_data: Optional[Mapping[str, str]] = None,
        **kwargs: Any,
    ) -> ChildOfRestObject:
        """Retrieve a single object.

        Args:
            id: The ID of the object to retrieve.
            include: Related data to include in the response.
            query_data: Dict of additional query parameters
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The response from the server, serialized into the object type.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """

        query_data = dict(query_data) if query_data else {}

        if include:
            query_data["include"] = ",".join(include)
        path = f"{self._path}/{id}"

        result = self.reporter.http_request(
            verb="get",
            path=path,
            query_data=query_data,
            **kwargs,
        )

        if TYPE_CHECKING:
            assert isinstance(self, RestManager)

        if hasattr(self, "_obj_cast"):
            return self._obj_cast(self.reporter, result.json())
        return self._obj_cls(self.reporter, result.json())


class GetRawMixin(Generic[ChildOfRestObject]):
    """Manager can retrieve raw file contents."""

    _path: str
    _obj_cls: Type[ChildOfRestObject]
    _obj_cast: Callable
    reporter: Reporter

    def get(
        self,
        id: str,
        **kwargs: Any,
    ) -> bytes:
        """Retrieve a single object.

        Args:
            id: Object ID
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The raw response data.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """

        path = f"{self._path}/{id}"

        return self.reporter.get_raw_file(path=path, **kwargs)


class _ListMixin(Generic[ChildOfRestObject]):
    """Parent class for ListMixin and SearchMixin."""

    _path: str
    _obj_cls: Type[ChildOfRestObject]
    _obj_cast: Callable
    reporter: Reporter

    def _get_list(  # pylint: disable = too-many-arguments, too-many-locals, redefined-builtin
        self,
        extra_path: str = "",
        term: Optional[str] = None,
        filter: Optional[Mapping[str, str]] = None,
        sort: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        query_data: Optional[Mapping[str, str]] = None,
        **kwargs: Any,
    ) -> RestList:
        """Retrieve a list of objects.

        Args:
            extra_path: Extra text to add to the request URL path
            term: A search term.
            filter: query string parameters for HTTP request of the form
                filter[field]
            sort: How to sort retrieved items
            include: Types of related data to include
            page: ID of the page to return - page[number]
            page_size: Number of items to return per page - page[size]
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            A :class:`~reporter.base.RestList` of :class:`~reporter.base.RestObject` instances.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """
        path = self._path + extra_path

        query_data = dict(query_data) if query_data else {}

        if term is not None:
            query_data["term"] = term

        if filter is None:
            filter = {}
        for key, value in filter.items():
            query_data[f"filter[{key}]"] = value

        if include:
            query_data["include"] = ",".join(include)

        if sort is not None and sort != []:
            query_data["sort"] = ",".join(sort)

        if page is not None:
            query_data["page[number]"] = str(page)

        if page_size is not None:
            query_data["page[size]"] = str(page_size)

        result = self.reporter.http_request(
            verb="get",
            path=path,
            query_data=query_data,
            **kwargs,
        )

        json = result.json()
        if TYPE_CHECKING:
            assert isinstance(self, RestManager)
        if hasattr(self, "_obj_cast"):
            data = [self._obj_cast(self.reporter, attrs) for attrs in json["data"]]
        else:
            data = [self._obj_cls(self.reporter, attrs) for attrs in json["data"]]
        links = json["links"]
        meta = json["meta"]

        return RestList(data=data, links=links, meta=meta)


class ListMixin(_ListMixin):
    """Manager can list objects."""

    def list(  # pylint: disable = too-many-arguments, redefined-builtin
        self,
        filter: Optional[Mapping[str, str]] = None,
        sort: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        query_data: Optional[Mapping[str, str]] = None,
        **kwargs: Any,
    ) -> RestList:
        """Retrieve a list of objects.

        Args:
            filter: query string parameters for HTTP request of the form
                filter[field]
            sort: How to sort retrieved items
            include: Types of related data to include
            page: ID of the page to return - page[number]
            page_size: Number of items to return per page - page[size]
            query_data: Dict of additional query parameters
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            A :class:`~reporter.base.RestList` of :class:`~reporter.base.RestObject` instances.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """
        return self._get_list(
            extra_path="",
            filter=filter,
            sort=sort,
            include=include,
            page=page,
            page_size=page_size,
            query_data=query_data,
            **kwargs,
        )


class SearchMixin(_ListMixin):
    """Manager can search for objects."""

    def search(  # pylint: disable = too-many-arguments
        self,
        term: Optional[str] = None,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        query_data: Optional[Mapping[str, str]] = None,
        **kwargs: Any,
    ) -> RestList:
        """Search for a list of objects.

        Args:
            term: Term to search for
            page: ID of the page to return - page[number]
            page_size: Number of items to return per page - page[size]
            query_data: Dict of additional query parameters
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            A :class:`~reporter.base.RestList` of :class:`~reporter.base.RestObject` instances.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """

        return self._get_list(
            extra_path="/search",
            page=page,
            page_size=page_size,
            term=term,
            query_data=query_data,
            **kwargs,
        )


class UpdateMixin(Generic[ChildOfRestObject]):
    """Manager can update objects."""

    _path: str
    _obj_cls: Type[ChildOfRestObject]
    _obj_cast: Callable
    reporter: Reporter

    def update(
        self,
        id: str,
        attrs: Mapping[str, Any],
        **kwargs: Any,
    ) -> ChildOfRestObject:
        """Update an object of type self._obj_cls.

        Args:
            id: ID of the object to update
            attrs: Attributes to update
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The response from the server, serialized into the object type.

        Raises:
            :class:`~reporter.exceptions.ReporterHttpError`: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """

        path = f"{self._path}/{id}"

        result = self.reporter.http_request(
            verb="patch",
            path=path,
            post_data=attrs,
            **kwargs,
        )

        if TYPE_CHECKING:
            assert isinstance(self, RestManager)
        if hasattr(self, "_obj_cast"):
            return self._obj_cast(self.reporter, result.json())
        return self._obj_cls(self.reporter, result.json())


class CrudMixin(
    CreateMixin, GetMixin, UpdateMixin, DeleteMixin, Generic[ChildOfRestObject]
):
    """Composite class of other mixins."""

    _obj_cls: Type[ChildOfRestObject]



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import ListMixin

__all__ = [
    "Language",
    "LanguageManager",
]


class Language(RestObject):
    pass


class LanguageManager(RestManager, ListMixin):
    _path = "languages"
    _obj_cls = Language



# pylint: disable = missing-module-docstring, missing-class-docstring

from typing import List
from reporter.base import RestManager, RestObject, Reporter
from reporter.mixins import CreateMixin, CrudMixin, DeleteMixin, ListMixin
from reporter.objects.task import Task


__all__ = [
    "TaskSet",
    "TaskSetManager",
    "AssessmentTaskSetManager",
]


class TaskSet(RestObject):
    pass


class TaskSetManager(RestManager, CrudMixin, ListMixin):
    _path = "task-sets"
    _obj_cls = TaskSet


class AssessmentTaskSetManager(RestManager, CreateMixin, DeleteMixin):
    _path = "assessments/{assessment_id}/task-sets"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = List[Task]

    def _obj_cast(self, reporter: Reporter, tasks: List):
        return [Task(reporter, task) for task in tasks]



"""RestObject _includes - see RestObject._deserialize_includes"""

# pylint: disable = wildcard-import, unused-wildcard-import, protected-access

from reporter.objects import *

Activity._includes = {
    "assessment": Assessment,
    "finding": Finding,
    "impersonator": User,
    "user": User,
}

Assessment._includes = {
    "activities": Activity,
    "assessmentTemplate": AssessmentTemplate,
    "assessmentUsers": AssessmentUser,
    "client": Client,
    "comments": AssessmentComment,
    "documents": Document,
    "findings": Finding,
    "language": Language,
    "nestedSections": AssessmentSection,
    "outputFiles": OutputFile,
    "phases": AssessmentPhase,
    "researchersOnReport": User,
    "sections": AssessmentSection,
    "targets": Target,
    "tasks": Task,
    "taskSets": TaskSet,
    "theme": Theme,
    "userGroups": UserGroup,
    "users": User,
}

AssessmentComment._includes = {
    "assessment": Assessment,
    "createdBy": User,
    "updatedBy": User,
    "documents": Document,
    "reactions": Reaction,
    "replies": AssessmentComment,
}

AssessmentPhase._includes = {
    "assessment": Assessment,
    "researchers": User,
    "reviewers": User,
}

AssessmentSection._includes = {
    "assessment": Assessment,
    "documents": User,
    "findings": Finding,
    "items": AssessmentSection,
    "comments": AssessmentSectionComment,
}

AssessmentSectionComment._includes = {
    "assessment": Assessment,
    "assessmentSection": AssessmentSection,
    "createdBy": User,
    "updatedBy": User,
    "documents": Document,
    "replies": AssessmentSectionComment,
}

AssessmentTemplate._includes = {
    "languages": Language,
    "taskSets": TaskSet,
}

AssessmentUser._includes = {
    "assessment": Assessment,
    "autoAssignments": AutoAssignment,
    "roles": Role,
    "user": User,
}

AutoAssignment._includes = {
    "assessment": Assessment,
    "user": User,
}

Client._includes = {
    "assessments": Assessment,
    "documents": Document,
    "userGroups": UserGroup,
    "users": User,
}

Document._includes = {
    "uploadedBy": User,
}

FindingComment._includes = {
    "assessment": Assessment,
    "finding": Finding,
    "createdBy": User,
    "updatedBy": User,
    "documents": Document,
    "reactions": Reaction,
    "replies": FindingComment,
}

FindingCreatedEvent._includes = {
    "assessment": Assessment,
    "finding": Finding,
    "createdBy": User,
    "reactions": Reaction,
    "replies": FindingComment,
}

FindingRetest._includes = {
    "assessments": Assessment,
    "finding": Finding,
    "createdBy": User,
    "updatedBy": User,
    "reviewedBy": User,
    "documents": Document,
    "reactions": Reaction,
    "retestInquiry": FindingRetestInquiry,
    "replies": FindingComment,
}

FindingRetestInquiry._includes = {
    "assessments": Assessment,
    "finding": Finding,
    "createdBy": User,
    "updatedBy": User,
    "convertedFromCommentBy": User,
    "documents": Document,
    "reactions": Reaction,
    "retest": FindingRetest,
    "replies": FindingComment,
}

FindingTemplate._includes = {
    "documents": Document,
}

Finding._includes = {
    "activities": Activity,
    "assessment": Assessment,
    "assessmentSection": AssessmentSection,
    "comments": FindingComment,
    "createdEvent": FindingCreatedEvent,
    "documents": Document,
    "resolvedTargets": Target,
    "resolvers": User,
    "retestInquiries": FindingRetestInquiry,
    "retests": FindingRetest,
    "targets": Target,
    "userGroups": UserGroup,
    "user": User,
}

Language._includes = {
    "assessments": Assessment,
}

OutputFile._includes = {
    "assessment": Assessment,
    "documents": Document,
}

Reaction._includes = {
    "assessment": Assessment,
    "user": User,
}

Target._includes = {
    "assessment": Assessment,
    "documents": Document,
    "findings": Finding,
}

Task._includes = {
    "assessment": Assessment,
    "client": Client,
    "completedBy": User,
    "finding": Finding,
    "taskSet": TaskSet,
    "users": User,
}

TaskSet._includes = {
    "assessments": Assessment,
    "assessmentTemplates": AssessmentTemplate,
    "copiedTasks": Task,
}

Theme._includes = {
    "assessments": Assessment,
    "assessmentTemplates": AssessmentTemplate,
    "documents": Document,
    "languages": Language,
}

User._includes = {
    "assessments": Assessment,
    "clients": Client,
    "documents": Document,
    "roles": Role,
    "tasks": Task,
    "userGroups": UserGroup,
}

UserGroup._includes = {
    "assessments": Assessment,
    "client": Client,
    "users": User,
    "findings": Finding,
}



"""This module exposes the :class:`~reporter.client.Reporter` object."""

import time
from typing import Any, Mapping, Optional, TYPE_CHECKING

import requests

import reporter.exceptions


__all__ = [
    "Reporter",
]


class Reporter:  # pylint: disable = too-many-instance-attributes, too-few-public-methods
    """Represents a Reporter server connection.

    Args:
        api_token: The Reporter API token to use for authentication.
        ssl_verify: Whether to verify the server's SSL certificate.
        url: The URL of the Reporter server. Must start with URL scheme (i.e. :code:`https://`).

    """

    api_token: str
    ssl_verify: bool
    url: str

    session: requests.Session
    """The ``requests.Session`` object used to make HTTP requests."""

    def __init__(
        self,
        api_token: str,
        url: str,
        ssl_verify: bool = True,
    ) -> None:
        """Initialize the Reporter instance.

        Args:
            api_token: The Reporter API token to use for authentication.
            ssl_verify: Whether to verify the server's SSL certificate.
            url: The URL of the Reporter server.

        """
        self.api_token = api_token
        self.ssl_verify = ssl_verify
        # Reporter does not accept double slash, but the user shouldn't be
        # expected to know that.
        self.url = url.rstrip("/")

        self.session = requests.Session()
        self.session.headers.update(
            {
                "Accept": "application/json",
                "Authorization": f"Bearer {api_token}",
            }
        )

        # Delay import until now to avoid circular import errors
        from reporter import (  # pylint: disable = import-outside-toplevel, cyclic-import
            objects,
        )

        self.activities = objects.ActivityManager(self)
        self.assessments = objects.AssessmentManager(self)
        self.assessment_comments = objects.AssessmentCommentManager(self)
        self.assessment_phases = objects.AssessmentPhaseManager(self)
        self.assessment_sections = objects.AssessmentSectionManager(self)
        self.assessment_section_comments = objects.AssessmentSectionCommentManager(self)
        self.assessment_templates = objects.AssessmentTemplateManager(self)
        self.clients = objects.ClientManager(self)
        self.documents = objects.DocumentManager(self)
        self.findings = objects.FindingManager(self)
        self.finding_comments = objects.FindingCommentManager(self)
        self.finding_retest_inquiries = objects.FindingRetestInquiryManager(self)
        self.finding_retests = objects.FindingRetestManager(self)
        self.finding_templates = objects.FindingTemplateManager(self)
        self.languages = objects.LanguageManager(self)
        self.output_files = objects.OutputFileManager(self)
        self.reactions = objects.ReactionManager(self)
        self.targets = objects.TargetManager(self)
        self.tasks = objects.TaskManager(self)
        self.task_sets = objects.TaskSetManager(self)
        self.themes = objects.ThemeManager(self)
        self.user_groups = objects.UserGroupManager(self)
        self.users = objects.UserManager(self)
        self.webhooks = objects.WebhookManager(self)

    def http_request(  # pylint: disable = too-many-arguments
        self,
        verb: str,
        path: str,
        headers: Optional[Mapping[str, str]] = None,
        query_data: Optional[Mapping[str, Any]] = None,
        post_data: Optional[Mapping[str, Any]] = None,
        files: Optional[Mapping[str, Any]] = None,
        obey_rate_limit=True,
    ) -> requests.Response:
        """Make an HTTP request to the Reporter server.

        Args:
            verb: The HTTP method to call (e.g. ``get``, ``post``, ``put``, ``delete``).
            path: Path to query (e.g. ``findings/1`` for ``/api/v1/findings/1``).
            headers: Extra HTTP headers; will overwrite default headers.
            query_data: Data to send as query string parameters.
            post_data: Data to send in the body. This will be converted to JSON unless
                ``files`` is not ``None``.
            files: The files to send in the request. If this is not ``None``, then the
                request will be a ``multipart/form-data`` request.
            obey_rate_limit: If ``True``, when receiving a 429 response, sleep
                for the amount of seconds specified in the response ``Retry-After``
                header before retrying the request.

        Returns:
            A requests Response object corresponding to the response from the Reporter
            server.

        Raises:
            ReporterHttpError: If the return code is not 2xx.
        """

        url = f"{self.url}/api/v1/{path}"

        # If files is present, we don't sent JSON.
        if files is not None:
            data = post_data
            json = None
        else:
            data = None
            json = post_data

        for i in range(2):
            result = self.session.request(  # pylint: disable = too-many-arguments
                method=verb,
                url=url,
                headers=headers,
                params=query_data,
                data=data,
                json=json,
                files=files,
                verify=self.ssl_verify,
            )
            if obey_rate_limit and result.status_code == 429 and i != 1:
                time.sleep(float(result.headers["Retry-After"]) + 0.5)
                continue
            break

        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)  # type: ignore

        if 200 <= result.status_code < 300:
            return result

        raise reporter.exceptions.ReporterHttpError(
            error_message=result.content,
            response_code=result.status_code,
            response_body=result.content,
        )

    def get_raw_file(
        self,
        path: str,
        headers=None,
        **kwargs,
    ) -> bytes:
        """Wrapper around :func:`http_request` for downloading a raw file as bytestring.

        Args:
            path: URL path of the raw file.
            headers: Request headers. Default: ``{"Accept": "*/*"}``.
            **kwargs: Extra options to pass to the underlying :func:`http_request` call.

        Returns:
            The raw file as a bytestring.
        """
        if headers is None:
            headers = {"Accept": "*/*"}

        result = self.http_request(
            verb="get",
            path=path,
            headers=headers,
            **kwargs,
        )

        return result.content



# pylint: disable = missing-module-docstring, missing-class-docstring

from typing import Any, Mapping, TYPE_CHECKING
from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, GetMixin, ListMixin, UpdateMixin

__all__ = [
    "Finding",
    "FindingManager",
    "AssessmentFindingManager",
]


class Finding(RestObject):
    pass


class FindingManager(RestManager, DeleteMixin, GetMixin, ListMixin, UpdateMixin):
    _path = "findings"
    _obj_cls = Finding


class AssessmentFindingManager(RestManager, CreateMixin):
    _path = "assessments/{assessment_id}/findings"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = Finding

    def create_from_template(
        self, template_id: str, attrs: Mapping[str, Any], **kwargs: Any
    ) -> Finding:
        """Create a new finding from a finding template

        Args:
            template_id: The ID of the finding template.
            attrs: Attributes for the created finding.
            kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The response from the server, serialized into the `Finding` type.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """
        if TYPE_CHECKING:
            assert isinstance(self._parent, RestObject)
        path = f"assessments/{self._parent.id}/finding-templates/{template_id}/findings"
        result = self.reporter.http_request(
            verb="post",
            path=path,
            post_data=attrs,
            files=None,
            **kwargs,
        )
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return Finding(self.reporter, result.json())



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin

__all__ = [
    "Reaction",
    "ReactionManager",
]


class Reaction(RestObject):
    pass


# In the Reporter API, the Reaction resources's DELETE method behaves unexpectedly.
# Instead of DELETE reactions/{id}, it is DELETE reactions, and the model info is given as body
# parameters. This will be fixed in the next release, and when python-reporter is updated we should
# add the DeleteMixin to this manager.
class ReactionManager(RestManager, CreateMixin):
    _path = "reactions"
    _obj_cls = Reaction



"""Reporter objects.

Instances of RestObject and RestManager corresponding to API models and their managers
respectively.

"""

from .activity import *
from .assessment import *
from .assessment_comment import *
from .assessment_phase import *
from .assessment_section import *
from .assessment_section_comment import *
from .assessment_section_event import *
from .assessment_template import *
from .assessment_user import *
from .auto_assignment import *
from .client import *
from .document import *
from .finding import *
from .finding_comment import *
from .finding_created_event import *
from .finding_event import *
from .finding_retest import *
from .finding_retest_inquiry import *
from .finding_template import *
from .language import *
from .output_file import *
from .reaction import *
from .role import *
from .target import *
from .task import *
from .task_set import *
from .theme import *
from .user import *
from .user_group import *
from .webhook import *



"""RestObject _includes - see RestObject.__init__"""

# pylint: disable = wildcard-import, unused-wildcard-import, protected-access

from reporter.objects import *

Assessment._children = {
    "comments": AssessmentAssessmentCommentManager,
    "findings": AssessmentFindingManager,
    "output_files": AssessmentOutputFileManager,
    "targets": AssessmentTargetManager,
    "tasks": AssessmentTaskManager,
    "task_sets": AssessmentTaskSetManager,
    "users": AssessmentAssessmentUserManager,
}

AssessmentComment._children = {
    "replies": AssessmentCommentReplyManager,
}

AssessmentSection._children = {
    "comments": AssessmentSectionAssessmentSectionCommentManager,
}

AssessmentSectionComment._children = {
    "replies": AssessmentSectionEventReplyManager,
}

Client._children = {
    "assessments": ClientAssessmentManager,
    "user_groups": ClientUserGroupManager,
}

Finding._children = {
    "comments": FindingFindingCommentManager,
    "retests": FindingFindingRetestManager,
    "retestInquiries": FindingFindingRetestInquiryManager,
}

FindingComment._children = {
    "replies": FindingEventReplyManager,
}

FindingRetest._children = {
    "replies": FindingEventReplyManager,
}

FindingRetestInquiry._children = {
    "replies": FindingEventReplyManager,
}



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import ListMixin

__all__ = [
    "Theme",
    "ThemeManager",
]


class Theme(RestObject):
    pass


class ThemeManager(RestManager, ListMixin):
    _path = "themes"
    _obj_cls = Theme



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, UpdateMixin

__all__ = [
    "AssessmentSectionComment",
    "AssessmentSectionCommentManager",
    "AssessmentSectionAssessmentSectionCommentManager",
]


class AssessmentSectionComment(RestObject):
    pass


class AssessmentSectionCommentManager(RestManager, UpdateMixin, DeleteMixin):
    _path = "assessment-section-comments"
    _obj_cls = AssessmentSectionComment


class AssessmentSectionAssessmentSectionCommentManager(RestManager, CreateMixin):
    _path = "assessment-sections/{assessment_section_id}/assessment-section-comments"
    _parent_attrs = {"assessment_section_id": "id"}
    _obj_cls = AssessmentSectionComment



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import ListMixin

__all__ = [
    "Role",
    "GlobalRoleManager",
    "AssessmentRoleManager",
]


class Role(RestObject):
    pass


class GlobalRoleManager(RestManager, ListMixin):
    _path = "roles"
    _obj_cls = Role


class AssessmentRoleManager(RestManager, ListMixin):
    _path = "assessment-roles"
    _obj_cls = Role
